# Copyright 2014 Canonical Ltd.  This software is licensed under the
# GNU General Public License version 3 (see the file LICENSE).
import os
import logging
import bzr
import phase
import shutil
import shlex
import re
import hashlib
import sys
import pwd
from .contain import (
    LXCContainer,
    NoSuchContainerError,
)
from .utils import (
    needs_root,
    mkdirs,
    mksymlink,
    CacheDisk,
)

MOJO_ROOT = "/srv/mojo"


#######################################
# Spec
#######################################

class SpecUninitializedError(Exception):
    pass


class ConfigNotFoundException(Exception):
    def __init__(self, message, search_list):
        self.message = message
        self.search_list = search_list


class Spec(object):
    def __init__(self, local_dir, remote=None, stage=None, manifest_file=None):
        self.local_dir = local_dir
        self.manifest_file = manifest_file
        self.local_branch = bzr.BzrBranch(self.local_dir)
        if remote:
            self.remote_branch = bzr.BzrBranch(remote)
        elif self.is_initialized:
            if self.local_branch.parent:
                self.remote_branch = bzr.BzrBranch(self.local_branch.parent)
            else:
                self.remote_branch = None
        else:
            raise ValueError("No remote or local branch specified for Spec")
        self.stage = stage
        self._manifest = None

    @property
    def is_initialized(self):
        return self.local_branch is not None and self.local_branch.is_branch

    def initialize(self):
        "Download or update spec on disk"
        if not os.path.isdir(self.local_dir):
            os.makedirs(self.local_dir)
        if self.local_branch.is_branch:
            self.local_branch.pull(src_location=self.remote_branch.location,
                                   overwrite=True)
        else:
            self.local_branch = self.remote_branch.clone(self.local_dir)
        self.local_branch.revert()
        self.local_branch.cleantree()

    @property
    def manifest(self):
        logging.info("Retrieve the spec's manifest")
        if self._manifest:
            return self._manifest
        if not self.is_initialized:
            return None
        self._manifest = Manifest(self, manifest_file=self.manifest_file,
            stage=self.stage)
        return self._manifest

    def _remove_duplicates(self, mylist):
        """Remove duplicates while preserving order.

        See http://stackoverflow.com/a/480227/523729
        """
        seen = set()
        seen_add = seen.add
        return [
            x for x in mylist
            if not (x in seen or seen_add(x))]

    def get_configs(self, base, stage=None):
        """Locate a list of stage-qualified configuration files.
           Configs are returned relative to the spec dir in this order:
             * <base>
             * <stage>/../<base>
             * <base>-<stage>
             * <stage>/<base>"""
        if not self.is_initialized:
            raise SpecUninitializedError("Spec has no local copy: {}"
                                         "".format(self.local_dir))
        config_names = [base]
        if stage:
            config_names.extend([
                "{}-{}".format(base, stage),
                os.path.join(stage, '../', base),
                os.path.join(stage, base),
            ])

            stage_dir = os.path.join(self.local_dir, stage)
            if os.path.exists(stage_dir):
                for subdir in os.listdir(stage_dir):
                    if subdir.startswith('.'):
                        continue
                    if os.path.isdir(os.path.join(stage_dir, subdir)):
                        config_names.extend([
                            os.path.join(stage, subdir, base)
                        ])

        configs = [
            config for config in config_names
            if os.path.exists(os.path.join(self.local_dir, config))]
        configs = [os.path.normpath(config) for config in configs]
        configs = self._remove_duplicates(configs)

        if configs:
            return configs

        raise ConfigNotFoundException("Config not found. Is MOJO_STAGE set "
                                      "or use the --stage option",
                                      config_names)

    def get_config(self, base, stage=None):
        """Locate a stage-qualified configuration file.
           The absolute path to the first of the following configs is returned:
             * <stage>/<base>
             * <base>-<stage>
             * <stage>/../<base>
             * <base>"""
        return os.path.join(self.local_dir, self.get_configs(base, stage)[-1])


#######################################
# Manifest
#######################################

class InvalidManifestException(Exception):
    pass


def parse_comment(bareline):
    text_comment = bareline[1:].strip()
    return "comment", dict(text=text_comment)


def parse_command(bareline):
    words = shlex.split(bareline)
    command = words[0]
    kwargs = {}
    for arg in words[1:]:
        k, v = arg.split("=", 1)
        kwargs[k] = v
    return command, kwargs


def parse_manifest(manifest_path):
    """Return a list of phases from the manifest file."""
    phases = []
    with open(manifest_path) as manifest_file:
        previous_phase = None
        for line in manifest_file:
            bareline = line.strip()
            if not bareline or bareline.startswith("##"):
                continue

            if bareline.startswith("#"):
                command, kwargs = parse_comment(bareline)
            else:
                command, kwargs = parse_command(bareline)

            if command == "include" and kwargs["config"]:
                parent_dir, _ = os.path.split(manifest_path)
                submanifest_path = os.path.join(parent_dir, kwargs['config'])
                phases.extend(parse_manifest(submanifest_path))
                continue

            try:
                current_phase = phase.build(command, **kwargs)
            except phase.NoSuchPhaseError as e:
                raise InvalidManifestException(e.message)

            if (previous_phase and previous_phase.name == "comment" and
                    current_phase.name == "comment"):
                previous_phase.options['text'] += (
                    "\n" + current_phase.options['text'])
            else:
                phases.append(current_phase)

            previous_phase = current_phase
    return phases


class Manifest(object):
    def __init__(self, spec, manifest_file, stage):
        super(Manifest, self).__init__()
        self.spec = spec
        self.config = spec.get_config(base=manifest_file, stage=stage)
        if manifest_file == "manifest":
            self.required_commands = ('collect', 'repo', 'deploy')
        else:
            self.required_commands = ()
        self.phases = []
        self.load()

    def load(self):
        self.phases = parse_manifest(self.config)
        commands = [phase.name for phase in self.phases]
        for command in self.required_commands:
            if command not in commands:
                raise InvalidManifestException("Missing required phase: {}"
                                               "".format(command))

    def run(self, project, workspace, series, stage=None):
        "Run all phases defined in this manifest"
        for phase in self.phases:
            phase.run(project, workspace, series, stage)


#######################################
# Monitor
#######################################

class Monitor(object):

    def __init__(self):
        self.CHANGED = 100

    def run(self, project, workspace, series, stage=None):
        """"Check for new revnos for branches
            Store the previous revnos in monitor dir
        """
        self.spec = workspace.spec
        self.monitor_dir = os.path.join(
            project.project_series, '{}.{}.{}'.format(
                workspace.name, stage, 'monitor'))
        mkdirs(self.monitor_dir)
        new = False
        with open(self.spec.get_config('collect', stage)) as config:
            for line in config:
                previous_revno = None
                if not line.strip().startswith("#") and not len(line.strip()) == 0:
                    options = {}
                    source = line.split()[1]
                    if ";" in source:
                        source, opt_string = source.strip().rsplit(";", 1)
                        for opt in opt_string.split(","):
                            options.update(dict((opt.split("="),)))
                    digest = hashlib.sha1(source).hexdigest()
                    branch = bzr.BzrBranch(source)
                    # Not a branch, could be a directory (@) or http source:
                    # ignore
                    if branch.is_branch:
                        current_revno = branch.revno
                        digestpath = os.path.join(self.monitor_dir, digest)
                        if os.path.exists(digestpath):
                            with open(digestpath) as digestfile:
                                for line in digestfile:
                                    previous_revno = line.split()[0]
                                    logging.debug("Read revno={} from {}: {}"
                                        "".format(previous_revno,
                                                  digestpath, line))
                        if not previous_revno == str(current_revno):
                            if "revno" in options:
                                logging.debug("Revno {} for branch {} has "
                                              "changed but configuration "
                                              "calls for revno {}"
                                              "".format(current_revno, source,
                                                        options["revno"]))
                                current_revno = options["revno"]
                            else:
                                new = True
                                logging.info("Revno {} for branch {} has "
                                             "changed from previous revno {}"
                                             "".format(current_revno, source,
                                                       previous_revno))
                            with open(digestpath, 'w') as digestfile:
                                digestfile.write("{} {}\n".format(
                                    current_revno, source))
        logging.info("Cache stats: {}".format(CacheDisk.get_stats()))
        if new:
            logging.info("Changes to branches for project: {}"
                         "".format(project.name))
            sys.exit(self.CHANGED)


#######################################
# Project
#######################################

class DuplicateProjectError(Exception):
    pass


class ProjectCreationError(Exception):
    pass


class Project(object):
    "A containerized mojo project"
    token_re = re.compile(r"[a-zA-Z0-9_.-]").search
    token_desc = "{} must only contain alphanumerics, underscores, or periods."

    def __init__(self, name, series, container_class=LXCContainer):
        if not name:
            raise ProjectCreationError("No project name specified. Use "
                                       "--project or set MOJO_PROJECT")
        if not bool(self.token_re(name)):
            raise ProjectCreationError(self.token_desc.format("Project names"))
        self.name = name
        self.series = series
        self.owner = os.environ.get('SUDO_USER')
        if not self.owner:
            # map my running uid to username
            self.owner = pwd.getpwuid(os.getuid()).pw_name

        ### Paths ###
        # /srv/mojo/<prj>
        self.root = Project.root_path(self.name)
        # /srv/mojo/<prj>/<series>
        self.project_series = os.path.join(self.root, self.series)
        # /srv/mojo/<prj>/<series>/ROOTFS
        self.container_root = os.path.join(self.project_series, "ROOTFS")
        # /srv/mojo/<prj>/<series>/ROOTFS/srv/mojo/<prj>/<series>
        self.lxc_project_series = self.container_root + self.project_series

        self.container = container_class(self.name, self.series)

    @classmethod
    def root_path(cls, name):
        return os.path.join(MOJO_ROOT, name)

    @classmethod
    def list_series(cls, name):
        return [s for s in sorted(os.listdir(cls.root_path(name)))]

    @classmethod
    def list(cls, only_names):
        "List all Mojo projects"
        proj_list = []
        for proj in os.listdir(MOJO_ROOT):
            if proj in ['LOCAL', 'bin', 'lib']:
                continue
            # Ignore any directories we don't have permissions to read in case
            # we have a multi-tenant mojo install. This is not expected as a
            # long term solution to multi-tenant mojo installs, as we should be
            # changing MOJO_ROOT for that, but it's not currently supported.
            try:
                os.listdir(os.path.join(MOJO_ROOT, proj))
            except OSError:
                continue
            if not only_names:
                series = " ".join(cls.list_series(proj))
                proj = "{}: {}".format(proj, series)
            proj_list.append(proj)
        return proj_list

    @needs_root
    def create(self):
        "Create a new LXC container and project"
        logging.info("Checking {} project path at {} ...".format(
            self.name, self.root))
        if os.path.exists(self.root):
            logging.warning("Project {} already exists at {}. Using existing "
                            "project".format(self.name, self.root))
        else:
            try:
                mkdirs(self.root, self.owner, mode=0700)
            except OSError as e:
                raise ProjectCreationError(
                    "Could not create project root: {}".format(e.strerror))

        if os.path.exists(self.project_series):
            logging.warning("Project {} project_series already exists at {}, "
                            "using it.".format(self.name, self.project_series))
        else:
            try:
                mkdirs(self.project_series, self.owner, mode=0700)
            except OSError as e:
                raise ProjectCreationError(
                    "Could not create project project_series: "
                    "{}".format(e.strerror))
        container_name = "{}.{}".format(self.name, self.series)
        lxc_root = "/var/lib/lxc/{}/rootfs".format(container_name)
        logging.info("Checking {} project LXC container at {} ..."
                     "".format(self.name, lxc_root))
        if not os.path.exists(lxc_root):
            logging.info("Not found, creating LXC container")
            # Can't use root location on precise's version of lxc
            # missing -B --dir option
            #self.container.create(self.container_root)
            self.container.create()

        logging.info("Checking {} project LXC container setup ..."
                     "".format(self.name))
        if not self.container.is_setup(lxc_root):
            logging.info("Not setup - setting up LXC container for mojo")
            self.container.setup()

        logging.info("Checking {} project container symlink {} -> {} ..."
                     "".format(self.name, lxc_root, self.container_root))
        if not os.path.lexists(self.container_root):
            os.symlink(lxc_root, self.container_root)

        logging.info("Checking {} project dir: {} ..."
                     "".format(self.name, self.lxc_project_series))
        if not os.path.exists(self.lxc_project_series):
            mkdirs(self.lxc_project_series, self.owner, mode=0700)

        sudoers_file = os.path.join(self.container_root, "etc", "sudoers.d",
                                    "ubuntu-all")
        logging.info("Writing {} sudoers file: {}".format(
                     self.name, sudoers_file))
        with open(sudoers_file, "w") as f:
            f.write("ubuntu ALL=(ALL) NOPASSWD: ALL\n")
        os.chmod(sudoers_file, 0400)

    def delete(self, force=True):
        """Delete a project and its LXC container - if no series remain,
        also delete project root"""
        try:
            shutil.rmtree(self.project_series)
        except OSError:
            if not force:
                raise
        try:
            self.container.destroy()
        except NoSuchContainerError:
            if not force:
                raise

        # If no series remain, delete project root
        if not Project.list_series(self.name):
            logging.info("Destroying project root: {}"
                         "".format(Project.root_path(self.name)))
            try:
                shutil.rmtree(Project.root_path(self.name))
            except OSError:
                if not force:
                    raise

    def ws_link(self, title):
        return os.path.join(self.project_series, str(title))

    def create_workspace(self, spec_url, title, stage=None,
                         manifest_file='manifest'):
        # If we're using nested directories for specs, make sure our workspace
        # symlink doesn't need additional sub-directories created.
        if "/" in title:
            title = title.replace("/", "-")
        ws_path = os.path.join(self.lxc_project_series, str(title))
        mkdirs(ws_path, self.owner)
        mksymlink(ws_path, self.ws_link(title))
        ws = Workspace(title, self.ws_link(title), stage=stage,
            manifest_file=manifest_file)
        ws.prepare()
        spec = Spec(ws.spec_dir, spec_url, stage=stage, manifest_file=manifest_file)
        spec.initialize()
        return ws

    def workspace(self, workspace_name):
        return Workspace(workspace_name, self.ws_link(workspace_name))

    def delete_workspace(self, workspace_name):
        ws = self.workspace(workspace_name)
        ws.delete()
        if os.path.lexists(self.ws_link(workspace_name)):
            os.unlink(self.ws_link(workspace_name))

    def workspace_list(self):
        return (w for w in os.listdir(self.lxc_project_series) if w != 'LOCAL')

    @property
    def utils(self):
        return self.container.utils


#######################################
# Workspace
#######################################

class NoSuchWorkspaceError(Exception):
    pass


class Workspace(object):
    "A mojo workspace"
    def __init__(self, name, path, stage=None, manifest_file=None):
        self.name = name
        self.manifest_file = manifest_file
        if not path:
            self.path = os.path.abspath(os.curdir)
            logging.info("No workspace path specified, using {}"
                         "".format(self.path))
        else:
            self.path = os.path.abspath(path)
        if not os.path.exists(self.path):
            raise NoSuchWorkspaceError(path)
        self.stage = stage

    @property
    def spec(self):
        if os.path.exists(self.spec_dir):
            spec_repo = bzr.BzrBranch(self.spec_dir)
            if spec_repo.is_branch:
                return Spec(local_dir=self.spec_dir, stage=self.stage,
                    manifest_file=self.manifest_file)
        return None

    @property
    def all_dirs(self):
        return [self.path,
                self.build_dir,
                self.local_dir,
                self.log_dir,
                self.spec_dir,
                ]

    def prepare(self):
        "Set up the workspace. Returns True on success, False on failure."
        for path in self.all_dirs:
            if not os.path.isdir(path):
                try:
                    os.makedirs(path)
                except OSError as e:
                    logging.error("{} ({})".format(e.strerror, e.filename))
                    return False
        return True

    def clean(self):
        "Reset the workspace"
        map(shutil.rmtree, (d for d in self.all_dirs if os.path.exists(d)))
        return self.prepare()

    def delete(self):
        "Destroy the workspace"
        realpath = os.path.realpath(self.path)
        shutil.rmtree(realpath)

    @property
    def build_dir(self):
        return os.path.join(self.path, "build")

    @property
    def repo_dir(self):
        return os.path.join(self.path, "charms")

    @property
    def local_dir(self):
        return os.path.join(self.path, "local")

    @property
    def log_dir(self):
        return os.path.join(self.path, "log")

    @property
    def spec_dir(self):
        return os.path.join(self.path, "spec")
