# Copyright 2014 Canonical Ltd.  This software is licensed under the
# GNU General Public License version 3 (see the file LICENSE).
from __future__ import print_function
import argparse
import logging
import sys
import os
import pkg_resources

import mojo

class SubcommandError(Exception):
    pass


def process_phaseopts(args):
    try:
        project = get_project(args, use_spec=False)
        series = get_series(args)
        workspace = get_workspace(args)
    except SubcommandError:
        raise ValueError()
    os.environ["MOJO_PROJECT"] = project
    os.environ["MOJO_WORKSPACE"] = workspace
    stage = get_stage(args)
    return project, workspace, stage, series, args.options


def action(name, args):
    try:
        project, workspace, stage, series, options = process_phaseopts(args)
    except ValueError:
        raise ValueError("Project, series and/or workspace not set. Are the "
                         "MOJO_PROJECT, MOJO_SERIES and MOJO_WORKSPACE env "
                         "variables set? Or the "
                         "--project/--series/--workspace options used?")
    workspace = mojo.Project(project, series).workspace(workspace)
    project = mojo.Project(project, series)
    phase = mojo.phase.build(name, **options)
    phase.run(project, workspace, series, stage)


class OptionParser(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        try:
            options = dict(option.split("=", 1) for option in values)
        except ValueError:
            parser.print_help(file=sys.stderr)
            sys.exit(1)
        setattr(namespace, self.dest, options)


# -----------------------
# helpers
# -----------------------

def get_series(args):
    """Return the series from args or the environment.

    Raises SubcommandError if no series can be determined.
    """
    series = args.series or os.environ.get("MOJO_SERIES")
    if not series:
        raise SubcommandError("No series given. Use --series or set "
                              "MOJO_SERIES")
    return series


def get_workspace(args):
    """Return the workspace from args or the environment.

    Raises SubcommandError if no workspace can be determined.
    """
    workspace = args.workspace or os.environ.get("MOJO_WORKSPACE")
    if not workspace:
        raise SubcommandError("No workspace given. Use --workspace or set "
                              "MOJO_WORKSPACE")
    return workspace


def get_project(args, use_spec=True):
    """Return the project from args or the environment.

    If use_spec, then the project is taken from the spec_url.
    Finally, the MOJO_PROJECT environment variable is used.

    Raises SubcommandError if no project can be determined.
    """
    project = args.project
    if not project and use_spec:
        project = args.spec_url.rsplit("/", 1)[-1]
    if not project:
        project = os.environ.get("MOJO_PROJECT")
    if not project:
        raise SubcommandError("No project given. Use --project or set "
                              "MOJO_PROJECT")
    return project


def get_stage(args):
    """Return the stage from args or the environment.

    Returns 'devel' by default.
    """
    return args.stage or os.environ.get("MOJO_STAGE") or 'devel'


def get_manifest_file(args):
    """Return the manifest file from args or default.

    We can't depend on default value of --manifest-file as it's
    only present for the run subcommand.
    """
    if hasattr(args, 'manifest'):
        return args.manifest

    return 'manifest'


# -----------------------
# commands
# -----------------------
def create_project(args):
    "Set up a Mojo project and LXC container"
    project_name = args.name
    series = get_series(args)
    project = mojo.Project(project_name, series)
    project.create()


def destroy_project(args):
    if not args.ffs:
        confirm = raw_input("Are you sure? ")
        if not confirm.strip().lower().startswith("y"):
            return
    if args.series:
        series_list = (args.series,)
    else:
        series_list = mojo.Project.list_series(args.name)
    for series in series_list:
        logging.info("Destroying project: {}, series: {}"
                     "".format(args.name, series))
        project = mojo.Project(args.name, series)
        project.delete()


def list_projects(args):
    map(print, mojo.Project.list(only_names=args.names))


def create_workspace(args):
    "Set up workspace and fetch a copy of the spec"
    project_name = get_project(args, use_spec=False)
    series = get_series(args)
    project = mojo.Project(project_name, series)
    stage = get_stage(args)
    manifest_file = get_manifest_file(args)

    try:
        ws = project.create_workspace(args.spec_url, args.name,
            stage=stage, manifest_file=manifest_file)
    except OSError as e:
        print("Are you sure the project exists?", file=sys.stderr)
        if args.blunt:
            raise
        sys.exit(2)
    logging.info("Workspace created. Validating manifest")
    try:
        manifest = ws.spec.manifest
        logging.info("Manifest OK ({})".format(manifest.config))
    except mojo.ConfigNotFoundException as e:
        logging.error(e.message)
        sys.exit(1)
    except ValueError as e:
        logging.error(e.message)
        sys.exit(1)


def destroy_workspace(args):
    if not args.ffs:
        confirm = raw_input("Are you sure? ")
        if not confirm.strip().lower().startswith("y"):
            return
    try:
        project_name = get_project(args, use_spec=False)
    except SubcommandError:
        raise SubcommandError("No project name given")
    series = get_series(args)
    mojo.Project(project_name, series).delete_workspace(args.name)


def list_workspaces(args):
    mojo_project = args.project or os.environ.get("MOJO_PROJECT")
    mojo_series = args.series or os.environ.get("MOJO_SERIES")
    if not mojo_project:
        raise SubcommandError("No project given. Use --project or set "
                              "MOJO_PROJECT")
    if not mojo_series:
        raise SubcommandError("No series given. Use --series or set "
                              "MOJO_SERIES")
    project = mojo.Project(mojo_project, mojo_series)
    map(print, project.workspace_list())


def builddeps(args):
    "Install build dependencies into project container"
    action('builddeps', args)


def collect(args):
    "Collect build resources using config-manager"
    action('collect', args)


def build(args):
    "Run deployment's build script"
    action('build', args)


def script(args):
    "Run deployment's build script"
    action('script', args)


def repo(args):
    "Build a charm repository"
    action('repo', args)


def sleep(args):
    "Sleep"
    action('sleep', args)


def secrets(args):
    "Secrets"
    action('secrets', args)


def deploy(args):
    "Build or update services"
    action('deploy', args)


def volumes(args):
    "Mount volumes"
    action('volumes', args)


def verify(args):
    "Verify the running service"
    action('verify', args)


def run_from_manifest(args):
    project_name = get_project(args)
    series = get_series(args)
    stage = get_stage(args)
    project = mojo.Project(project_name, series)
    workspace = project.create_workspace(args.spec_url,
        args.name, stage=stage, manifest_file=args.manifest_file)
    manifest = workspace.spec.manifest
    manifest.run(project, workspace, series, stage)


def monitor_specification(args):
    project_name = args.project or args.spec_url.rsplit("/", 1)[-1]
    project = mojo.Project(project_name, args.series)
    workspace = project.create_workspace(args.spec_url, args.name,
        stage=args.stage, manifest_file=args.manifest_file)
    monitor = mojo.Monitor()
    monitor.run(project, workspace, args.series, args.stage)


def create_arg_parser():
    # main program parser
    ap = argparse.ArgumentParser()
    ap.add_argument("--blunt", action="store_true",
                    help="Handle fewer exceptions")
    ap.add_argument('--version', action='version',
                    version=pkg_resources.require("mojo")[0].version)

    # subcommand parser
    sp = ap.add_subparsers(title="subcommands")

    # options parser for phase commands
    phaseopts = argparse.ArgumentParser(add_help=False)
    phaseopts.add_argument("--stage", help="Deployment stage (default: devel)")
    phaseopts.add_argument("--options",
                           help="Phase options. i.e. config=CONFIG_FILENAME",
                           nargs="*", metavar="key=value", action=OptionParser,
                           default={})
    phaseopts.add_argument("-p", "--project",
                           help="Project name. Overrides the MOJO_PROJECT "
                           "environment variable.")
    phaseopts.add_argument("-w", "--workspace", help="Workspacename. Overrides"
                           "the MOJO_WORKSPACE environment variable.")
    phaseopts.add_argument("-s", "--series", help="Distro series. Overrides"
                           "the MOJO_SERIES environment variable.")
    phaseopts.add_argument("-l", "--log_level", default='INFO',
                           help="Log level. Default to INFO")

    # project-new subcommand - creates a new mojo project
    prjnewcmd = sp.add_parser("project-new",
                              description="Create a new Mojo project and "
                              "container")
    prjnewcmd.add_argument("-s", "--series", help="Distro series",
                           required=False)
    prjnewcmd.add_argument("name",
                           help="Base name of the project ([a-zA-Z0-9._]*)")
    prjnewcmd.set_defaults(func=create_project)

    # project-destroy subcommand - deletes a mojo project
    prjdelcmd = sp.add_parser("project-destroy",
                              description="Create a new Mojo project and "
                              "container")
    prjdelcmd.add_argument("name",
                           help="Base name of the project ([a-zA-Z0-9._]*)")
    prjdelcmd.add_argument("-y", "--ffs", action="store_true",
                           help="Don't ask for confirmation")
    prjdelcmd.add_argument("-s", "--series", help="Distro series "
                           "(default=<all found>)")
    prjdelcmd.set_defaults(func=destroy_project)

    # project-list
    prjlistcmd = sp.add_parser("project-list",
                               description="List existing projects")
    prjlistcmd.add_argument("-n", "--names", action="store_true",
                            help="List only project names, without series")
    prjlistcmd.set_defaults(func=list_projects)

    # workspace-new subcommand - creates a new workspace for a specific spec
    # revision
    wsnewcmd = sp.add_parser("workspace-new",
                             description="Initialize a mojo workspace")
    wsnewcmd.add_argument("spec_url", help="Configuration spec url (bzr)")
    wsnewcmd.add_argument("name",
                          help="Name of the workspace (suggested: revno)")
    wsnewcmd.add_argument("-p", "--project",
                          help="Project name (equivalent to MOJO_PROJECT "
                          "environment variable)")
    wsnewcmd.add_argument("-s", "--series", help="Distro series. Overrides"
                          "the MOJO_SERIES environment variable.")
    wsnewcmd.add_argument("--stage", default="devel",
                          help="Deployment stage. Defaults to 'devel'")
    wsnewcmd.set_defaults(func=create_workspace, parser=wsnewcmd)

    # workspace-destroy subcommand - deletes a mojo project
    wsdelcmd = sp.add_parser("workspace-destroy",
                             description="Create a new Mojo project and "
                             "container")
    wsdelcmd.add_argument("-p", "--project",
                          help="Project name (equivalent to MOJO_PROJECT "
                          "environment variable)")
    wsdelcmd.add_argument("-s", "--series", help="Distro series. Overrides"
                          "the MOJO_SERIES environment variable.")
    wsdelcmd.add_argument("name", help="Workspace name")
    wsdelcmd.add_argument("-y", "--ffs", action="store_true",
                          help="Don't ask for confirmation")
    wsdelcmd.set_defaults(func=destroy_workspace)

    # workspace-list subcommand
    wslistcmd = sp.add_parser("workspace-list",
                              description="List workspaces for a project")
    wslistcmd.add_argument("-p", "--project",
                           help="Project name (equivalent to MOJO_PROJECT "
                           "environment variable)")
    wslistcmd.add_argument("-s", "--series", help="Distro series. Overrides"
                           "the MOJO_SERIES environment variable.")
    wslistcmd.set_defaults(func=list_workspaces)

    # builddeps subcommand - install build dependencies for project
    depscmd = sp.add_parser("builddeps", description="builddeps",
                            parents=[phaseopts])
    depscmd.set_defaults(func=builddeps)

    # collect subcommand - collects code branches to build directory
    clctcmd = sp.add_parser("collect",
                            description="Collect resources for build",
                            parents=[phaseopts])
    clctcmd.set_defaults(func=collect)

    # build subcommand - run project build script inside container
    buildcmd = sp.add_parser("build",
                             description="Execute project build script",
                             parents=[phaseopts])
    buildcmd.set_defaults(func=build)

    # repo subcommand - assembles charms into a repository
    repocmd = sp.add_parser("repo", description="Assemble a charm repository",
                            parents=[phaseopts])
    repocmd.set_defaults(func=repo)

    # sleep subcommand - sleeps for n seconds
    sleepcmd = sp.add_parser("sleep", description="Sleep for time in seconds",
                             parents=[phaseopts])
    sleepcmd.set_defaults(func=sleep)

    # secrets subcommand - secretss for n seconds
    secretscmd = sp.add_parser("secrets",
                               description="Copy secrets from staged "
                               "location to workspace local",
                               parents=[phaseopts])
    secretscmd.set_defaults(func=secrets)

    # script subcommand - run an arbitrary script inside container
    scriptcmd = sp.add_parser("script", description="Execute a project script",
                              parents=[phaseopts])
    scriptcmd.set_defaults(func=script)

    # deploy subcommand - run a juju-deployer session
    deploycmd = sp.add_parser("deploy",
                              description="Deploy services and/or create "
                              "relations",
                              parents=[phaseopts])
    deploycmd.add_argument("--ignore", action="store_true", default=False,
                           help="Exit 0 even if service deployment fails")
    deploycmd.set_defaults(func=deploy)

    # volumes subcommand - mount volumes
    volcmd = sp.add_parser("volumes", description="Mount volumes",
                           parents=[phaseopts])
    volcmd.set_defaults(func=volumes)

    # verify subcommand - verify that a deployment was successful
    verifycmd = sp.add_parser("verify",
                              description="Verify that the deployment is "
                              "working",
                              parents=[phaseopts])
    verifycmd.set_defaults(func=verify)

    # run subcommand - run an entire spec per its manifest
    runcmd = sp.add_parser("run",
                           description="Run an entire deployment from start "
                           "to finish")
    runcmd.add_argument("spec_url", help="Bzr URL to the Mojo spec")
    runcmd.add_argument("name",
                        help="Name of the workspace (suggested: revno)")
    runcmd.add_argument("--manifest-file", default="manifest",
                        help="The particular manifest file to run")
    runcmd.add_argument("--series", help="Distro series. Overrides"
                        "the MOJO_SERIES environment variable.")
    runcmd.add_argument("--stage",
                        help="Deployment stage. Defaults to 'devel'")
    runcmd.add_argument("--project",
                        help="Name for the project (defaults to spec name)")
    runcmd.add_argument("-l", "--log_level", default='INFO',
                        help="Log level. " "Default to INFO")
    runcmd.set_defaults(func=run_from_manifest)

    # monitor subcommand - monitor specification collect file for
    # updates
    monitorcmd = sp.add_parser("monitor",
                               description="Monitor specification collect "
                               "file for updates")
    monitorcmd.add_argument("spec_url", help="Bzr URL to the Mojo spec")
    monitorcmd.add_argument("name",
                            help="Name of the workspace (suggested: revno)")
    monitorcmd.add_argument("--stage", default="devel",
                            help="Deployment stage. Defaults to 'devel'")
    monitorcmd.add_argument("--project",
                            help="Name for the project (defaults to spec "
                            "name)")
    monitorcmd.add_argument("--series", help="Distro series. Overrides"
                            "the MOJO_SERIES environment variable.")
    monitorcmd.add_argument("--cache_ttl", type=int, default=3600,
                            help="Number of seconds to cache bzr revnos"
                            "(default=%s)" % 3600)
    monitorcmd.add_argument("-l", "--log_level", default='INFO',
                            help="Log level. " "Default to INFO")
    monitorcmd.set_defaults(func=monitor_specification)

    return ap


def run_with_args(args, ap=None):
    if not "log_level" in args:
        args.log_level = "INFO"
    logging.basicConfig(level=args.log_level,
                        format="%(asctime)s [%(levelname)s] %(message)s",
                        datefmt="%Y-%m-%d %H:%M:%S")
    if hasattr(args, "cache_ttl") and args.cache_ttl is not None:
        mojo.CacheDisk.ttl(args.cache_ttl)
    try:
        args.func(args)
    except ValueError as e:
        print(e.message, file=sys.stderr)
        if ap is not None:
            ap.print_help(file=sys.stderr)
        sys.exit(1)
    except mojo.utils.Unprivileged:
        print("This sub-command must be run as root", file=sys.stderr)
        sys.exit(1)
    except mojo.ConfigNotFoundException as e:
        print(e.message, file=sys.stderr)
        print("Searched:", file=sys.stderr)
        for path in e.search_list:
            print("  {}".format(path), file=sys.stderr)
        sys.exit(1)


# -----------------------
# main
# -----------------------


def main():
    ap = create_arg_parser()
    args = ap.parse_args()
    run_with_args(args, ap)
