# Copyright 2014 Canonical Ltd.  This software is licensed under the
# GNU General Public License version 3 (see the file LICENSE).
#
# check for LXC container for workspace
# create LXC container for workspace
# execute scripts inside of container
# enable or disable networking inside the container

import os
import subprocess
from contextlib import contextmanager
import shlex
import shutil
import logging
from .utils import needs_root
import yaml
import re
import distutils.spawn
import pwd


class NoSuchContainerError(Exception):
    pass


class InvalidConfigurationError(Exception):
    pass


# Decorator for methods that require a defined container
def ensure_defined(method):
    def wrapped(self, *args, **kwargs):
        if not self.defined:
            raise NoSuchContainerError(self.container_name)
        return method(self, *args, **kwargs)
    return wrapped


class Container(object):
    def __init__(self, name, series, ensure_exists=False):
        self.name = name
        self.series = series
        self.container_name = "{}.{}".format(name, series)
        if ensure_exists and not self.defined:
            raise NoSuchContainerError(self.container_name)
        if self.defined:
            self.load_config()
        self.utils = ContainerUtils(self)

    def load_config(self):
        raise NotImplemented

    @contextmanager
    def running_container(self, network=True, keep_running=False):
        running = self.running
        if not running:
            self.start(network=network)
        try:
            yield
        finally:
            if not running and not keep_running:
                self.stop()

    def create(self):
        raise NotImplemented

    def destroy(self, force=False):
        raise NotImplemented

    def start(self):
        raise NotImplemented

    def stop(self):
        raise NotImplemented

    def run(self, command, env=None, user=None, network=True, stdout=None,
            stderr=None):
        raise NotImplemented

    @property
    def defined(self):
        return False

    @property
    def running(self):
        return False

    @property
    def container_path(self):
        return None

    @property
    def rootfs(self):
        return None


class LXCContainer(Container):
    LXC_DIR = "/var/lib/lxc"

    @ensure_defined
    def load_config(self):
        self.config = {}
        is_config_line = lambda l: l.strip() and not l.strip().startswith("#")
        with open(os.path.join(self.LXC_DIR, self.container_name, "config")) as cf:
            # ignore comments
            for line in cf:
                if is_config_line(line):
                    k, v = line.split("=", 1)
                    self.config[k.strip()] = v.strip()

    @needs_root
    def create(self, root_location=None):
        if self.defined:
            raise InvalidConfigurationError("Duplicate container: {}"
                                            "".format(self.container_name))
        # we want to use the ubuntu template so that the environment is
        # consistent with what users expect, but we also make our own
        # adjustments to it as part of Project setup. It might be worth
        # building our own LXC template for Mojo at some point.
        cmd = ("lxc-create",
               "-t", "ubuntu",
               "-n", self.container_name,
               )
        if root_location is not None and os.path.isdir(os.path.dirname(
                                                       root_location)):
            cmd += ("-B", "dir", "--dir", root_location)
        elif root_location:
            raise InvalidConfigurationError(
                "Root location must be a directory")
        cmd += ("--",)
        if self.series is not None:
            cmd += ("--release", self.series)
        logging.info("Creating LXC container '{}'".format(self.container_name))
        logging.info("Running: '{}'".format(cmd))
        subprocess.check_call(cmd)
        logging.info("LXC creation Complete")

    @needs_root
    def setup(self, root_location=None):
        pkgs_to_install = ['software-properties-common',
                           'python-software-properties',
                           'rsync']
        # lxc < 1.0 and the precise kernel do not have a feature needed to use
        # lxc-start, lxc-execute requires that lxc be installed inside the lxc
        # container.
        if not self.lxc_1_0():
            pkgs_to_install.append('lxc')
        logging.info("Bootstrapping lxc env")
        self.load_config()
        # Serge Hallyn's solution to bootstrapping uses custom init scripts
        # The first init script runs whenever the lxc is started and it updates
        # and installs lxc and other packages needed
        with open(os.path.join(self.rootfs, 'etc', 'init',
                  'containersetup.conf'),
                  'w') as containersetup:
            containersetup.write(
                '''description "container setup job"
author "Serge Hallyn <serge.hallyn@canonical.com>"

start on runlevel [2345]
script
   apt-get update
   apt-get -y dist-upgrade
   apt-get -y install {}
end script\n'''.format(" ".join(pkgs_to_install)))
        # The second init script runs after the first has completed and
        # stops the lxc signaling completion of the installs
        # It also stops itself from running again by echoing 'manual'
        # into itself
        with open(os.path.join(self.rootfs, 'etc', 'init',
                  'setupstopped.conf'), 'w') as setupstopped:
            setupstopped.write(
                '''description "stop lxc containersetup has stopped"
author "Serge Hallyn <serge.hallyn@canonical.com>"
start on stopped containersetup

script
echo manual >> /etc/init/setup-stopped.conf
/sbin/poweroff
end script\n''')
        # Starting the lxc will run the init scripts
        self.start()
        # Wait for the lxc to stop signaling the completion of the installs
        self.wait("STOPPED")
        logging.info("Finished Bootstrapping")

        logging.info("Creating network-disabled and nested lxc configurations")
        shutil.copy(self.conf, self.conf_virtnet)
        with open(self.conf_virtnet) as virtnet:
            with open(self.conf_nonet, "w") as nonet, open(self.conf,
                                                           "w") as hostnet:
                for line in virtnet:
                    # Handle cgroups in a nested lxc per Serge Hallyn
                    line = line.replace(
                        "#lxc.aa_profile = lxc-container-default-with-nesting",
                        "lxc.aa_profile = lxc-container-default-with-nesting")
                    if not line.strip().startswith("lxc.network."):
                        hostnet.write(line)
                        nonet.write(line)
                nonet.write("lxc.network.type=empty\n")
        self.load_config()
        logging.info("LXC Setup Complete")

    def is_setup(self, root_location=None):
        setup_file = root_location + "/etc/init/setup-stopped.conf"
        return os.path.exists(setup_file)

    @needs_root
    @ensure_defined
    def destroy(self, force=False):
        cmd = ("lxc-destroy", "-n", self.container_name)
        if force:
            cmd += ("-f",)
        subprocess.check_output(cmd)

    @needs_root
    @ensure_defined
    def add_mount(self, host_path, guest_path=None):
        logging.info("Adding host mount to LXC container '{}'"
                     "".format(self.container_name))
        if not os.path.exists(host_path):
            raise InvalidConfigurationError("Host path {} does not exist."
                                            "".format(host_path))
        if not guest_path:
            dest_path = self.rootfs + host_path
        else:
            dest_path = self.rootfs + guest_path
        os.makedirs(dest_path)
        fstab_line = "{} {} none bind 0 0".format(host_path, dest_path)
        subprocess.check_output("echo {} >> {}".format(fstab_line, self.fstab),
                                shell=True)

    @needs_root
    @ensure_defined
    def remove_mount(self, host_path):
        logging.info("Removing host mount from LXC container '{}'"
                     "".format(self.container_name))
        cmd = "sed --in-place -e '#{}#d' {}".format(host_path, self.fstab)
        subprocess.check_output(cmd, shell=True)

    def check_state(self, state=None):
        if state:
            cmd = ('sudo', 'lxc-info', '-n', self.container_name)
            containers = yaml.safe_load(subprocess.check_output(cmd))
            # lxc-info shows 'state' (precise) or 'State' (newer)
            return state in containers.get('state', containers.get('State', {}))
        cmd = ('sudo', 'lxc-ls', '-1')
        containers = subprocess.check_output(cmd).split()
        return self.container_name in containers

    @ensure_defined
    def start(self, network=True):
        if self.running:
            logging.info("LXC container '{}' is already running"
                         "".format(self.container_name))
        else:
            logging.info("Starting LXC container '{}'".format(
                self.container_name))
            # always start containers in the background
            cmd = ["sudo", "lxc-start", "-d", "-n", self.container_name]
            keys = {}
            if not network:
                cmd += ["-f", self.conf_nonet]
            cmd += ["-s {}={}".format(k, v) for k, v in keys.items()]
            logging.debug("running {}".format(cmd))
            subprocess.check_call(cmd)
            self.wait("RUNNING")

    @ensure_defined
    def stop(self, wait=True):
        if not self.running:
            logging.info("LXC container '{}' is already stopped"
                         "".format(self.container_name))
        else:
            logging.info("Stopping LXC to container '{}'".format(
                self.container_name))
            cmd = ("sudo", "lxc-shutdown")
            if wait:
                cmd += ("-w",)
            cmd += ("-n", self.container_name)
            subprocess.check_call(cmd)

    @ensure_defined
    def wait(self, state='RUNNING'):
        cmd = ("sudo", "lxc-wait", "-n", self.container_name, "-s", state)
        logging.info("Waiting for LXC container to get to state '{}'"
                     "".format(state))
        subprocess.check_call(cmd)

    def lxc_1_0(self):
        # lxc-version command is missing at (since?) 1.0.0beta3,
        # adapt until it's addressed at:
        # https://bugs.launchpad.net/ubuntu/+source/lxc/+bug/1273689
        if distutils.spawn.find_executable('lxc-version'):
            (_, _, version) = subprocess.check_output(
                ['lxc-version']).rstrip().split(' ')
        else:
            version = subprocess.check_output(
                ['lxc-cgroup', '--version']).rstrip()

        # 1.0 features, since 1.0.0.betaX (but not alphaX):
        # --share-net (doesn't require installing LXC in the container itself
        return not re.match(r'^(1\.0\.0\.alpha|0\.)', version)

    @ensure_defined
    def run(self, command, env=None, user=None, network=True, stdout=None,
            stderr=None):
        "Run commands in the LXC environment"
        if not command:
            logging.warn("Refusing to execute empty command")
            return -1
        if not user:
            if os.environ.get('SUDO_USER'):
                user = os.environ.get('SUDO_USER')
            else:
                user = pwd.getpwuid(os.getuid()).pw_name
        # Avoid looping fork bomb by not running when called as user
        # root
        if user != 'root':
            if not self.utils.user_exists(user):
                self.utils.create_user(user)
        command = ["sudo", "-E", "-u", user] + shlex.split(command)
        if isinstance(env, dict):
            env_vars = ["env"] + ["=".join((k, v)) for k, v in env.items()]
            command = env_vars + command
        if self.lxc_1_0():
            cmd = ["sudo", "lxc-start", "-n", self.container_name, "-F"]
            if network:
                cmd.extend(["--share-net", "1"])
        else:
            cmd = ["sudo", "lxc-execute", "-n", self.container_name]
            if not network:
                cmd.extend(["-f", self.conf_nonet])
        cmd.extend(["--"])
        cmd.extend(command)
        logging.info("Running command in container '{}': {}"
                     "".format(self.container_name, ' '.join(command)))
        return subprocess.check_call(cmd, stdout=stdout, stderr=stderr)

    @property
    def defined(self):
        return self.check_state()

    @property
    def running(self):
        return self.check_state("RUNNING")

    @property
    def container_path(self):
        return os.path.join(self.LXC_DIR, self.container_name)

    @property
    @ensure_defined
    def rootfs(self):
        return self.config['lxc.rootfs']

    @property
    def conf(self):
        return os.path.join(self.container_path, "config")

    @property
    def conf_virtnet(self):
        return os.path.join(self.container_path, "config.virtnet")

    @property
    def conf_nonet(self):
        return os.path.join(self.container_path, "config.nonet")

    @property
    def fstab(self):
        return os.path.join(self.container_path, "fstab")


class ContainerUtils(object):
    def __init__(self, container):
        super(ContainerUtils, self).__init__()
        self.container = container

    def add_repo_key(self, repo_key):
        command = 'apt-key add {}'.format(repo_key)
        self.container.run(command, user="root")

    def add_repo(self, repo_spec, confirm=True):
        if confirm:
            confirmation = "-y"
        else:
            confirmation = ""
        command = 'add-apt-repository {} "{}"'.format(confirmation, repo_spec)
        self.container.run(command, user="root")
        self.apt_update()

    def apt_update(self):
        with open(os.devnull, 'w') as devnull:
            self.container.run("env DEBIAN_FRONTEND=noninteractive "
                               "apt-get update", user="root", stdout=devnull)
            self.container.run("env DEBIAN_FRONTEND=noninteractive "
                               "apt-get -y dist-upgrade", user="root",
                               stdout=devnull)

    def install_packages(self, *packages, **kwargs):
        prompt = kwargs.get("prompt", False)
        if prompt:
            assume_yes = ""
        else:
            assume_yes = "-y"
        command = "apt-get {} install {}" "".format(assume_yes,
                                                    " ".join(packages))
        self.container.run(command, user="root")

    def user_exists(self, user):
        command = "getent passwd {}".format(user)
        try:
            with open(os.devnull, 'w') as devnull:
                self.container.run(command, user="root", stdout=devnull)
            return True
        except:
            return False

    def create_user(self, user):
        (username, shadow, uid,
         gid, name, home, shell) = subprocess.check_output(
            ['getent', 'passwd', user]).split(':')
        # if the uid exists update the name to match the caller
        try:
            command = "useradd --home {} --create-home --uid {} {}" \
                "".format(home, uid, username)
            self.container.run(command, user="root")
        except subprocess.CalledProcessError as e:
            if e.returncode == 4:
                command = "cp /etc/sudoers.d/ubuntu-all " \
                    "/etc/sudoers.d/{}-all".format(username)
                self.container.run(command, user="root")
                command = 'sed -i "s/ubuntu/{}/g" ' \
                    '/etc/sudoers.d/{}-all'.format(username, username)
                self.container.run(command, user="root")
                command = 'usermod -l {} -G sudo ubuntu'.format(username)
                self.container.run(command, user="root")
            else:
                raise e
