# Copyright 2014 Canonical Ltd.  This software is licensed under the
# GNU General Public License version 3 (see the file LICENSE).
from __future__ import print_function
import argparse
import yaml
from subprocess import check_output
import logging
import sys


class Status(object):
    def __init__(self):
        cmd = ['juju', 'status']
        output = check_output(cmd)
        self.status = yaml.safe_load(output)

    def _service_info(self, svc):
        if svc == "ALL":
            units = {}
            for service in self.status['services'].values():
                for unit, data in service.get('units', {}).items():
                    units[unit] = data
        elif svc not in self.status['services']:
            return None
        else:
            service = self.status['services'][svc]
            units = service.get('units', {})
        return units

    def service_ips(self, service):
        units = self._service_info(service)
        if units:
            return [u.get('public-address') for u in units.values()
                    if u.get('public-address')]
        return units

    def service_units(self, service):
        units = self._service_info(service)
        if units:
            return units.keys()
        return units

    def service_machine_numbers(self, service):
        units = self._service_info(service)
        if units:
            return [u.get('machine') for u in units.values()
                    if u.get('machine')]
        return units

    def service_instance_ids(self, service):
        machine_numbers = self.service_machine_numbers(service)
        if machine_numbers is None:
            return None
        elif not machine_numbers:
            return []
        return [self.machine_machine_id(m)[0] for m in machine_numbers]

    def machine(self, machine_num):
        if machine_num not in self.status['machines']:
            return None
        return self.status['machines'][machine_num]

    def machine_ip(self, machine_num):
        machine = self.machine(machine_num)
        if machine is None:
            return None
        return machine.get('dns-name')

    def machine_machine_id(self, machine_num):
        machine = self.machine(machine_num)
        if machine is None:
            return None
        return machine.get('instance-id')


# -----------------------
# main
# -----------------------


def main():
    description = """
    List properties for objects in a Juju environment.
    Available properties:

    unit
       servicename
         ips
         machine_numbers
         instance_ids
    machine
       machine_number
         ips
         instance_ids"""
    ap = argparse.ArgumentParser(description=description)

    ap.add_argument("type", nargs=1, help="Object type (unit, machine)")
    ap.add_argument("inst", nargs=1, help="Object instance")
    ap.add_argument("prop", nargs=1, help="Property to list")

    args = ap.parse_args()

    objtype = args.type[0]
    inst = args.inst[0]
    prop = args.prop[0]

    attr = "{}_{}".format(objtype, prop)
    status = Status()
    handler = getattr(status, attr)
    if not handler:
        logging.error("Can't get {} for a {}".format(prop, objtype))
    values = handler(inst)

    if values is None:
        logging.error("{} not found: {}".format(objtype.capitalize(), inst))
        sys.exit(1)
    elif not values:
        logging.warning("{} {} has no {}".format(objtype.capitalize(), inst,
                        prop))
    else:
        if isinstance(values, list):
            map(print, values)
        else:
            print(values)
    sys.exit(0)
