# Copyright 2014 Canonical Ltd.  This software is licensed under the
# GNU General Public License version 3 (see the file LICENSE).
import os
import logging
import subprocess
import charm_repo
import contain
import re
import shutil
import time
from .utils import (
    bicommand,
    chdir,
    get_dep_prog,
)
from jinja2 import Template
import mojo


class NoSuchPhaseError(Exception):
    pass


class FileDoesNotExistException(Exception):
    pass


class NonNumericException(Exception):
    pass


class JujuDeployerException(Exception):
    pass


class Phase(object):
    "Base class for phases"
    name = "not_implemented"

    def __init__(self, **kwargs):
        super(Phase, self).__init__()
        # check for config= then script= then name of phase
        self.config_base = kwargs.get('config',
                                      kwargs.get('script', self.name))
        self.options = kwargs

    def __str__(self):
        return "<{} {} {} {}>".format(self.__class__.__name__, self.name,
                                      self.config_base, self.options)

    def run(self, project, workspace, series, stage=None):
        raise NotImplementedError


class SleepPhase(Phase):
    """Sleep for time in seconds
       Arguments:
           config      number of seconds to sleep
                       Default: 30
    """
    name = "sleep"

    def run(self, project, workspace, series, stage=None):
        "Sleep for time in seconds"
        logging.debug("### Running phase {} with options {} ###"
                      "".format(self.name, self.options))
        if self.config_base == 'sleep':
            self.config_base = 30
        logging.info("Sleeping for {} seconds".format(self.config_base))
        try:
            self.config_base = int(self.config_base)
            time.sleep(self.config_base)
        except:
            raise NonNumericException("Time for sleep phase must be the "
                                      "numeric number of seconds not '{}'"
                                      "".format(self.config_base))


class CommentPhase(Phase):
    """Include a well-deliniated comment in the info logger.

    When writing a spec, comments in a manifest are useful not only when
    reading the manifest, but also in the output when running the manifest.

    For convenience, any line(s) in a manifest beginning with a single
    hash will be converted to a comment. If you require a comment in
    the manifest which isn't useful during running, use a double-hash.

    This phase cannot be run from the command line.

    Arguments:
        text      The text to be displayed.
                  Default: "comment"
    """
    name = "comment"

    def run(self, *args, **kwargs):
        border = "#" * 77
        comment = "\n".join([border, self.options['text'], border])
        logging.info("Manifest comment:\n\n" + comment + "\n\n")


class SecretsPhase(Phase):
    """Pull Secrets into workspace local
       Arguments:
           config      Directory where secrets are located
                       Default: /srv/mojo/LOCAL/$MOJO_PROJECT/$MOJO_STAGE
    """
    name = "secrets"

    def run(self, project, workspace, series, stage=None):
        "Pull Secrets into workspace local"
        logging.debug("### Running phase {} with options {} ###"
                      "".format(self.name, self.options))
        if self.config_base == 'secrets':
            self.config_base = os.path.join(mojo.MOJO_ROOT, 'LOCAL',
                                            project.name, stage or 'devel')
        logging.info("Pulling secrets from {} to {}"
                     "".format(self.config_base, workspace.local_dir))
        if os.path.isdir(self.config_base):
            src_files = os.listdir(self.config_base)
            for file_name in src_files:
                full_file_name = os.path.join(self.config_base, file_name)
                if os.path.isfile(full_file_name):
                    shutil.copy(full_file_name, workspace.local_dir)
                elif os.path.isdir(full_file_name):
                    dest_dir = os.path.join(workspace.local_dir, file_name)
                    if os.path.exists(dest_dir):
                        shutil.rmtree(dest_dir)
                    shutil.copytree(full_file_name,
                                    os.path.join(workspace.local_dir,
                                                 file_name))
                else:
                    raise FileDoesNotExistException("{} is not a file or "
                                                    "directory"
                                                    "".format(full_file_name))
        else:
            raise FileDoesNotExistException("Secrets directory {} does not "
                                            "exist.".format(self.config_base))


class CollectPhase(Phase):
    """Collect code and resources into workspace build directory.
       Arguments:
           config      base name for config-manager file
    """
    name = "collect"

    def run(self, project, workspace, series, stage=None):
        "Collect build resources using config-manager"
        logging.debug("### Running phase {} with options {} ###"
                      "".format(self.name, self.options))
        logging.info("Building resource tree")
        codetree = get_dep_prog("codetree")
        config = workspace.spec.get_config(self.config_base, stage)
        cmd = [codetree, config]
        with chdir(workspace.build_dir):
            subprocess.check_call(cmd)


class InstallBuildDepsPhase(Phase):
    """Install dependencies into project container
       Arguments:
           repo_keys   comma-separated list of apt repository gpg key
                       file names
           repos       comma-separated list of apt repositories
           packages    comma-separated list of packages to insall
    """
    name = "builddeps"

    def run(self, project, workspace, series, stage=None):
        "Install packages"
        logging.debug("### Running phase {} with options {}"
                      "".format(self.name, self.options))
        logging.info("Installing apt repos and packages")
        utils = project.utils
        repo_keys = self.options.get('repo_keys')
        if repo_keys:
            for repo_key in repo_keys.split(","):
                repo_key = os.path.join(workspace.spec_dir, stage or 'devel',
                                        repo_key)
                utils.add_repo_key(repo_key)
        repos = self.options.get('repos')
        if repos:
            for repo in repos.split(","):
                utils.add_repo(repo)
        packages = self.options.get('packages')
        if packages:
            utils.apt_update()
            utils.install_packages(*packages.split(","))


class VolumesPhase(Phase):
    """Attach persistent volumes to services.
       Arguments:
           config    base name for volumes config file
    """
    name = "volumes"

    def run(self, project, workspace, series, stage=None):
        "Attach volumes to services"
        logging.debug("### Running phase {} with options {}".format(self.name,
                      self.options))
        logging.info("Mounting volumes")
        mojovol = get_dep_prog("mojo-vol")
        config = workspace.spec.get_config(self.config_base, stage)
        cmd = [mojovol, '--config', config]
        subprocess.check_call(cmd)


class ScriptPhase(Phase):
    """Execute a script within the build directory.
       Arguments:
           config      base name for script file
           lxc         Boolean to use lxc or not
           KEY=VALUE   set as environment variables for use by the script
    """
    name = "script"

    def run(self, project, workspace, series, stage=None, lxc=False, network=True):

        logging.debug("### Running phase {} with options {}".format(self.name,
                      self.options))
        logging.info("Running script {}".format(self.config_base))
        # Override use of lxc from options
        if "lxc" in self.options:
            lxc = str2bool(self.options['lxc'])

        "Run a script portion of a deployment process"
        # XXX: Run all scripts in a chroot or LXC environment
        env = {}
        # Set any options to the environment
        for key in self.options:
            env[key] = self.options[key]
        env['MOJO_PROJECT'] = project.name
        env['MOJO_WORKSPACE'] = workspace.name
        env['MOJO_SERIES'] = series
        env['MOJO_WORKSPACE_DIR'] = workspace.path
        env['MOJO_BUILD_DIR'] = workspace.build_dir
        env['MOJO_REPO_DIR'] = workspace.repo_dir
        env['MOJO_SPEC_DIR'] = workspace.spec_dir
        env['MOJO_LOG_DIR'] = workspace.log_dir
        env['MOJO_LOCAL_DIR'] = workspace.local_dir
        env['MOJO_STAGE'] = stage or 'devel'
        script = workspace.spec.get_config(self.config_base, stage)
        if lxc:
            lxc = contain.LXCContainer(project.name, series, ensure_exists=True)
            lxc.run(script, env=env, network=network)
        else:
            env_vars = ["env"] + ["=".join((k, v)) for k, v in env.items()]
            command = env_vars + [script]
            subprocess.check_call(command)


class BuildPhase(ScriptPhase):
    """Execute a build script within a build directory. The build environment
       is restricted to have no access to networks.
       Arguments:
           config     Base name for build script file
    """
    name = "build"

    def run(self, project, workspace, series, stage):
        "Run a build script"
        # XXX: Remove network access during build phase
        super(BuildPhase, self).run(project, workspace, series, stage,
                                    lxc=True, network=False)


class CharmRepoPhase(Phase):
    """Build a charm repo from paths within the build directory.
       Arguments:
           config    Base name for the charm repo spec
    """
    name = "repo"

    def run(self, project, workspace, series, stage):
        "Build a charm repository"
        logging.debug("### Running phase {} with options {}"
                      "".format(self.name, self.options))
        logging.info("Build a charm repository")
        repo = charm_repo.CharmRepo(workspace.repo_dir)
        config = workspace.spec.get_config(self.config_base, stage)
        repo.build_from_config(config, workspace.build_dir, series)


class DeployerPhase(Phase):
    """Run juju-deployer.
       Arguments:
           config         Base name for deployer config
           optional=y/N   When deployment fails, ignore it
           local          Base name in local dir for deployer config (secrets)
           juju           Juju environment
           delay          Deploy Delay in seconds
                          Default: 60 seconds
           timeout        Overall timeout for the deployment
                          Default: 2700 seconds (45 minutes)
    """
    name = "deploy"

    def run(self, project, workspace, series, stage):
        logging.debug("### Running phase {} with options {}"
                      "".format(self.name, self.options))
        logging.info("Running juju-deployer")
        deployer = get_dep_prog("juju-deployer")

        # Process the deployment configs as a Jinja2 template
        tmpl_vars = {
            'workspace': workspace,
            'series': series,
            'spec_dir': workspace.spec_dir,
            'local_dir': workspace.local_dir,
            'build_dir': workspace.build_dir,
            'stage': stage,
        }
        configs = ()
        for config_src in workspace.spec.get_configs(self.config_base, stage):
            # XXX Why render configs to the juju repo dir?
            config_dest = os.path.join(workspace.repo_dir, config_src)
            if not os.path.exists(os.path.dirname(config_dest)):
                os.makedirs(os.path.dirname(config_dest))
            self.render_config(
                os.path.join(workspace.spec_dir, config_src), config_dest,
                tmpl_vars)
            configs += ('-c', config_dest)

        if "local" in self.options:
            local_cfg_src = os.path.join(workspace.local_dir,
                                         self.options["local"])
            if not os.path.exists(local_cfg_src):
                raise FileDoesNotExistException("Local config file does not "
                                                "exist: {}"
                                                "".format(local_cfg_src))
            local_cfg = os.path.join(workspace.repo_dir, "local.cfg")
            self.render_config(local_cfg_src, local_cfg, tmpl_vars)
            configs += ('-c', local_cfg)
        if "juju" in self.options:
            configs += ('-e', self.options['juju'])
        elif os.environ.get("JUJU_ENV"):
            configs += ('-e', os.environ.get("JUJU_ENV"))

        if "delay" in self.options:
            deploy_delay = ('-s', str(self.options.get("delay")))
        else:
            deploy_delay = ('-s', '60')

        if "timeout" in self.options:
            timeout = ('-t', str(self.options.get("timeout")))
        else:
            timeout = ()

        if "target" in self.options:
            deploy_name = self.options.get("target")
        else:
            # get the (first, and assumed only) deployment name
            # from the config file
            cmd = ['python', deployer, '-c', configs[1], "-l"]
            deploy_name = subprocess.check_output(cmd).split("\n")[0]

        # run the deploy
        env = os.environ.copy()
        env['MOJO_SERIES'] = series
        env['MOJO_STAGE'] = stage or 'devel'
        cmd = ('python', deployer)
        cmd += configs
        cmd += deploy_delay
        cmd += timeout
        cmd += ('-d', deploy_name, '-W', '-u')
        if str2bool(self.options.get('optional')):
            with chdir(workspace.repo_dir):
                subprocess.call(cmd)
        else:
            print("Running: {}".format(" ".join(cmd)))
            with chdir(workspace.repo_dir):
                status, output = bicommand(" ".join(cmd), showoutput=True, env=env)
            if status != 0:
                try:
                    fail = re.search(".*unit: (?P<f>[^:]+):.*hook failed",
                        output).group("f")
                    print("Deployment failed, grabbing the last 200 lines of "
                          "output from the juju logs on {}".format(fail))
                    cmd = ["juju", "ssh", fail, "tail", "-200",
                        "/var/log/juju/unit-%s.log" % (
                        fail.replace("/", "-"),)]
                    with chdir(workspace.repo_dir):
                        subprocess.call(cmd)
                except AttributeError:
                    # We didn't find a match on the fail string
                    pass
                finally:
                    print("Status: {}".format(status))
                    raise JujuDeployerException

    def render_config(self, src, dest, tmpl_vars=None):
        if not tmpl_vars:
            tmpl_vars = {}
        with open(src) as inp:
            with open(dest, "w") as outp:
                outp.write(Template(inp.read()).render(**tmpl_vars))


class VerifyPhase(ScriptPhase):
    """Execute a script to verify environment"
       Arguments:
           config      base name for script file
           lxc         Boolean to use lxc or not
    """
    name = "verify"

    def run(self, project, workspace, series, stage):
        super(VerifyPhase, self).run(project, workspace, series, stage)


# ========================================
# Phase creation helper
# ========================================

def phase_types(phase_class=None):
    "Build a dict of phases and their subtypes"
    if phase_class is None:
        phase_class = Phase
    types = {phase_class.name: phase_class}
    for subclass in phase_class.__subclasses__():
        types.update(phase_types(subclass))
    return types


def build(phase_name, **kwargs):
    "Build a phase by name, passing all arguments as a single string"
    if getattr(phase_types, "cache", None) is None:
        phase_types.cache = phase_types(Phase)
    if phase_name not in phase_types.cache:
        raise NoSuchPhaseError("{} is not a registered phase type"
                               "".format(phase_name))
    return phase_types.cache[phase_name](**kwargs)


def str2bool(v):
    if v:
        return v.lower() in ("yes", "true", "t", "1")
    return False
