# Copyright 2014 Canonical Ltd.  This software is licensed under the
# GNU General Public License version 3 (see the file LICENSE).

from unittest import TestCase

from mojo import Manifest
from mojo.tests.utils import MojoTestCaseMixin


test_manifests_for_includes = [{
    'name': "An include at the end of a file",
    'expected_phases': ['collect', 'repo', 'deploy', 'script', 'verify'],
    'manifests': {
        'manifest': """
collect
repo
deploy config=services
include config=other-manifest
""",
        'other-manifest': """
script config=post-deploy
verify
""",
    },
}, {
    'name': "An include in the middle of a file",
    'expected_phases': ['collect', 'repo', 'deploy', 'script', 'verify'],
    'manifests': {
        'manifest': """
collect
repo
deploy config=services
include config=other-manifest
verify
""",
        'other-manifest': """
script config=post-deploy
""",
    }
}, {
    'name': "Nested include files",
    'expected_phases': [
        'collect',  # manifest
        'repo',
        'deploy',   # deploy-manifest
        'verify',   # verify-manifest
        'script',
        'verify',
        'deploy',   # deploy-manifest
        'script',   # manifest
    ],
    'manifests': {
        'manifest': """
collect
repo
include config=deploy-manifest
script
""",
        'deploy-manifest': """
deploy config=services
include config=verify-manifest
deploy config=somethingelse
""",
        'verify-manifest': """
verify
script config=other-test
verify config=verify-nagios
""",
    },
}]


class ManifestTestCase(TestCase, MojoTestCaseMixin):

    def test_include_sets_correct_phases(self):
        for example in test_manifests_for_includes:
            spec = self.create_spec(manifests=example['manifests'])

            manifest = Manifest(spec=spec, manifest_file='manifest',
                                stage='devel')

            actual_phases = [phase.name for phase in manifest.phases]
            self.assertEqual(
                example['expected_phases'], actual_phases,
                "The example '{}' expected the phases:\n{}\n "
                "but had\n{}\ninstead.".format(
                    example['name'], example['expected_phases'],
                    actual_phases))

    def test_manifest_comment_parsing(self):
        manifest_text = """
collect
## Should be ignored.
repo
# Here we go
deploy config=services
# Done
# with a multiline comment.
script config=post-deploy
# Verify
verify
"""
        spec = self.create_spec(manifests=dict(manifest=manifest_text))
        manifest = Manifest(spec=spec, manifest_file='manifest',
                            stage='devel')

        actual_phases = [phase.name for phase in manifest.phases]
        self.assertEqual([
            'collect',
            'repo',
            'comment',
            'deploy',
            'comment',
            'script',
            'comment',
            'verify',
        ], actual_phases)
        self.assertEqual("Here we go", manifest.phases[2].options['text'])
        self.assertEqual("Done\nwith a multiline comment.",
                         manifest.phases[4].options['text'])
        self.assertEqual("Verify", manifest.phases[6].options['text'])
